from shapely.geometry import shape
import geopandas as gpd
import pandas as pd
from elasticsearch import Elasticsearch

class Document(object):
    def __init__(self, es, index, doc_type):
        self.documents = self.read_index(es, index, doc_type)

    def read_index(self, es, index, doc_type):
        data = {}
        size = 1000
        has_max = hasattr(self, 'max') and self.max is not None
        if has_max:
            if self.max < size:
                size = self.max

        print('Query on ES database for fields:')
        print(self.get_query())
        res = es.search(index=index, doc_type=doc_type, scroll='60s', size=size,
                        body=self.get_query())
        while len(res['hits']['hits']) > 0 and (not has_max or len(data) < self.max):
            data.update({ hit['_id']: hit['_source'] for hit in res['hits']['hits']})
            while True:
                try:
                    res = es.scroll(res['_scroll_id'], scroll='60s')
                    break
                except:
                    continue

        return data

    def get_query(self):
        return {'query': {'match_all': {}}}

class Fields(Document):
    def __init__(self, es, index, doc_type, years=None, max=1000):
        self.years = years
        self.max = max
        super(Fields, self).__init__(es, index, doc_type)

    @staticmethod
    def __clean_coordinates(poly):
        coords = []

        for coord in poly:
            _ls = []
            for ls in coord:
                _ls.append(ls[:2])
            coords.append(_ls)

        return coords

    @staticmethod
    def __clean_geometry(geom):
        if geom is None:
            return None

        if geom['type'] == 'MultiPolygon':
            geom['coordinates'] = [Fields.__clean_coordinates(poly) for poly in geom['coordinates']]
        elif geom['type'] == 'Polygon':
            geom['coordinates'] = Fields.__clean_coordinates(geom['coordinates'])

        return geom

    def dataframe(self):
        return None

    def get_query(self):
        if self.years is not None:
            return {
                'query': {
                    'constant_score' : {
                        'filter' : {
                            'term' : { 'metadata.year' : self.years[0]}
                        }
                    }
                }
            }
        else:
            return super(Fields,self).get_query()

class FieldDataWIGFl(Fields):
    def __init__(self, es, index, doc_type, years=None, max=1000, parcelid=None):
        self.parcelid = parcelid
        super(FieldDataWIGFl, self).__init__(es, index, doc_type, years, max)

    @staticmethod
    def __clean_coordinates(poly):
        coords = []

        for coord in poly:
            _ls = []
            for ls in coord:
                _ls.append(ls[:2])
            coords.append(_ls)

        return coords

    @staticmethod
    def __clean_geometry(geom):
        if geom is None:
            return None

        if geom['type'] == 'MultiPolygon':
            geom['coordinates'] = \
              [FieldDataWIGFl.__clean_coordinates(poly) for poly in geom['coordinates']]
        elif geom['type'] == 'Polygon':
            geom['coordinates'] = FieldDataWIGFl.__clean_coordinates(geom['coordinates'])

        return geom

    # Overwrite dataframe method to take into account databio specific ES database structure.
    def dataframe(self):
        docs = []
        for doc_id in self.documents:
            doc = self.documents[doc_id]
            d = {}

            if 'cropFenology' in doc:
                d = doc['cropFenology']
                d['VARIETY'] = doc.get('cropFenology', {}).get('variety')

            # Cleanup fields that don't fit well in dataframe.
            if 'desiccation' in d:
                del d['desiccation']
            d['field_id'] = doc_id
            if 'metadata' in doc:
                d['geometry'] = self.__clean_geometry(doc.get('metadata', {}).get('geometry'))
                d['year'] = doc.get('metadata', {}).get('year')
                d['farmer'] = doc.get('metadata', {}).get('farmer')
                d['fieldName'] = doc.get('metadata', {}).get('localFieldName')

                if 'geometry' not in d:
                    d['geometry'] = self.__clean_geometry(doc.get('geometry', {}))

                docs.append(d)

        if len(docs) > 0:
            return pd.DataFrame(docs).set_index('field_id')
        else:
            print('NO FIELDS FOR THIS PARCEL ID')
            return pd.DataFrame(docs)

    def get_query(self):
        if  self.parcelid is not None:
            return {
                'query': {
                    'constant_score': {
                        'filter': {
                            'term': {
                                '_id': self.parcelid
                            }
                        }
                    }
                }
            }

        else:
            return super(Fields, self).get_query()

class ParcelWIGVlaanderen:
    def __init__(self, db_url='http://es-01.vgt.vito.be:9200', index='watchitgrow', type='field'):
        self.db_url = db_url
        self.index = index
        self.type = type
        self._years = None
        self._max = None
        self._parcelid = None

        self.inputdescription = {
            'years':{
                'type': 'int',
                'collectiontype': 'List',
                'required' : True
            },
            'max': {
                'type': 'int',
                'required' : False
            }
        }

    def years(self, *years):
        if years[0] is not None:
            self._years = years
        return self

    def max(self,max):
        self._max = max
        return self

    def parcelid(self, parcelid):
        self._parcelid = parcelid
        return self

    # Overwrite read function to take into account datastructures in the databio ES database.
    def read(self):
        es = Elasticsearch(hosts=self.db_url)

        dataframe = FieldDataWIGFl(es, self.index, self.type, self._years, max=self._max,
                                   parcelid=self._parcelid).dataframe()

        if len(dataframe) == 0:
            return dataframe

        # clean up fields with no geometry
        dataframe = dataframe[~dataframe['geometry'].isnull()]

        dataframe['geometry'] = dataframe['geometry'].map(shape)
        crs = {'init': 'epsg:4326'}
        dataframe = gpd.GeoDataFrame(dataframe, crs=crs, geometry=dataframe['geometry'])

        return dataframe

def parcel_wig(year=None, id=None, db_url='http://es-01.vgt.vito.be:9200', index='watchitgrow_7'):
    parcel = ParcelWIGVlaanderen(db_url=db_url, index=index, type='field').years(year).parcelid(id)


    return parcel
